;;; patcher-mail.el --- Mail management

;; Copyright (C) 2008, 2009, 2010, 2011 Didier Verna
;; Copyright (C) 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2007 Didier Verna

;; Author:        Didier Verna <didier@xemacs.org>
;; Maintainer:    Didier Verna <didier@xemacs.org>
;; Created:       Mon Feb 15 15:26:26 2010
;; Last Revision: Thu Dec  8 08:53:32 2011
;; Keywords:      maint


;; This file is part of Patcher.

;; Patcher is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License version 3,
;; as published by the Free Software Foundation.

;; Patcher is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program; if not, write to the Free Software
;; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


;;; Commentary:

;; Contents management by FCM version 0.1.


;;; Code:

(require 'cl)

(eval-when-compile (require 'patcher-cutil))
(require 'patcher-util)
(require 'patcher-project)
(require 'patcher-instance)
(require 'patcher-source)
(require 'patcher-change-log)
(require 'patcher-diff)
(require 'patcher-cmtcmd)
(require 'patcher-logmsg)


;; Require 'sendmail for getting `mail-header-separator'.
;; #### Now that a fake mail sending function exists, sendmail shouldn't be
;; #### systematically required like this.  However, since most users will
;; #### really want do send real messages, it probably doesn't hurt to keep
;; #### things as-is.
(require 'sendmail)



;; ==========================================================================
;; Utilities
;; ==========================================================================

(make-variable-buffer-local
 (defvar patcher-change-logs-marker nil
   ;; Marker indicating the beginning of the ChangeLog entries, when they are
   ;; separated from the patch.
   ))

(make-variable-buffer-local
 (defvar patcher-diff-marker nil
   ;; Marker indicating the beginning of the diff.
   ))

(defun patcher-goto-subject ()
  ;; Move point to the beginning of the Subject: header's contents.
  (goto-char (point-min))
  (re-search-forward "^Subject: "))

(defun patcher-goto-signature ()
  ;; Move point to the beginning of the mail signature (actually, in front of
  ;; the signature separator), if any.  Otherwise, move point to the end of
  ;; the message. Return that position.
  (goto-char (point-min))
  (if (re-search-forward
       (cond ((eq major-mode 'mail-mode)
	      ;; this is hard-wired in sendmail.el
	      "\n\n-- \n")
	     ((eq major-mode 'message-mode)
	      (declare-boundp message-signature-separator))
	     (t
	      (patcher-warning "\
Major mode: %s.
Your mailing method is not fully supported by Patcher.
This is not critical though: Patcher may not find the message signature
correctly.

Please report to <didier@xemacs.org>."
			       major-mode)
	      ;; Use the standard one by default.
	      "\n\n-- \n"))
       nil t)
      (goto-char (match-beginning 0))
    ;; else: no signature
    (goto-char (point-max))))




;; ==========================================================================
;; Diff generation
;; ==========================================================================

(globally-declare-boundp 'unused)

(defun* patcher-diff-project
    (project
     &aux (buffer (patcher-project-process-buffer project))
	  (updating (patcher-project-option project :change-logs-updating))
	  (appearance (patcher-project-option project
			:change-logs-appearance)))
  ;; (Re)generate PROJECT's diff and insert items in the current mail buffer.
  ;; Depending on PROJECT settings (ChangeLogs updating mode and appearance),
  ;; insertion of ChangeLog entries or even the diff itself might be delayed.
  (patcher-condition-case condition
      ;; Note that in case of a diff regeneration, PROJECT may contain old
      ;; source or ChangeLog references, and also maybe retained ChangeLog
      ;; skeletons. We need to deal with that by doing some additional cleanup
      ;; (like forgetting about obsolete source or ChangeLog files) or some
      ;; additional work (like generating newly required skeletons).
      (multiple-value-bind (sources change-logs)
	  (patcher-diff-specification project)
	;; First, figure out what's new and what's obsolete. When we end up
	;; with obsolete source or ChangeLog files, it's like they never
	;; existed, so there's no point in honoring the :kill-*-after-sending
	;; options.
	(multiple-value-bind (unused obsolete-sources new-sources)
	    (patcher-list= (patcher-project-sources project) sources
	      :test #'string=)
	  (patcher-unlink-sources project
	    :sources obsolete-sources
	    :override-kill t)
	  (patcher-link-sources project new-sources))
	(multiple-value-bind (unused obsolete-change-logs new-change-logs)
	    (patcher-list= (patcher-project-change-logs project) change-logs
	      :test #'string=)
	  ;; Having obsolete ChangeLogs that still contain skeletons is very
	  ;; likely to be an error.
	  (when (eq updating 'automatic)
	    (let (generated-change-log-buffers)
	      (dolist (obsolete-change-log obsolete-change-logs)
		(multiple-value-bind (obsolete-change-log-buffer)
		    (patcher-change-log-buffer project obsolete-change-log)
		  (when (and obsolete-change-log-buffer
			     (patcher-change-log-extent project
			       obsolete-change-log-buffer))
		    (patcher-endpush obsolete-change-log-buffer
				     generated-change-log-buffers))))
	      (when generated-change-log-buffers
		(case (patcher-with-message
			  (format "\
WARNING: some ChangeLog skeletons remain in obsolete ChangeLog files (not
involved with the project anymore). The relevant ChangeLog files are the
following: %s.

By answering the question below, you have the possibility to keep those
skeletons (n), remove them all (y) or choose interactively (i).

Note that if you keep some of them, you may run into trouble later..."
			      (patcher-buffers-string
			       generated-change-log-buffers))
			(patcher-read-char
			 "Remove obsolete ChangeLog skeletons? " "yni"))
		  (?y (patcher-ungenerate-change-logs project
			generated-change-log-buffers))
		  (?i (patcher-ungenerate-change-logs project
			generated-change-log-buffers
			:interactive t
			:prompt "Remove this obsolete skeleton? "))))))
	  (patcher-unlink-change-logs project
	    :change-logs obsolete-change-logs
	    :override-kill t)
	  (patcher-link-change-logs project new-change-logs)
	  ;; Now, insert whatever possible in the mail buffer, perform some
	  ;; sanity checks based on what we know and generate ChangeLog
	  ;; skeletons if needed.
	  (case updating
	    ;; No ChangeLog. The diff we already have is good enough so we can
	    ;; insert it right now.
	    ((nil)
	     (patcher-insert-diff project patcher-diff-marker)
	     (patcher-message "\
To commit your changes, type \\[patcher-mail-commit]."))
	    ;; Automatic mode. ChangeLogs insertion, if required, is postponed
	    ;; but we might still be able to insert the diff right now.
	    (automatic
	     ;; New ChangeLogs must be clean. Otherwise the project is out of
	     ;; date. Note that except for new ChangeLogs, the current diff
	     ;; might still contain ChangeLog entries: obsolete ones that were
	     ;; generated and not removed, and correct ones that were not
	     ;; ungenerated.
	     (patcher-detect-spurious-change-logs project new-change-logs)
	     ;; When ChangeLogs are inserted in a separate place, we can
	     ;; already insert the source diff.
	     (when (member appearance '(verbatim pack nil))
	       (patcher-insert-diff project patcher-diff-marker))
	     ;; Finally, we need to generate the missing skeletons. Since
	     ;; patch-to-change-log doesn't know if some of them have been
	     ;; generated already, I need to remove from the process buffer
	     ;; all source diffs that would contribute to an already generated
	     ;; skeleton. I definitely need to write my own version of
	     ;; patch-to-change-log.
	     ;; #### WARNING: it seems unsafe to delete extents from within
	     ;; mapcar-extents (I got cases where the mapping were interrupted
	     ;; before all extents were processed).
	     (dolist (extent (patcher-source-extents buffer))
	       (multiple-value-bind (change-log)
		   (patcher-change-log-buffer project
		       (patcher-locate-change-log project
			 (extent-property extent 'patcher-source)))
		 (when (and change-log
			    (patcher-change-log-extent project change-log))
		   (patcher-delete-extent-and-region extent))))
	     (patcher-generate-change-logs project)
	     (patcher-message "Please annotate the ChangeLog skeletons%s."
			      (case appearance
				(verbatim
				 "\
, and type \\[patcher-mail-insert-change-logs] to insert them")
				(pack
				 "\
, and type \\[patcher-mail-insert-change-logs] to diff them")
				(patch "\
, and type \\[patcher-mail-insert-change-logs] to create the whole diff")
				((nil)
				 (if (patcher-project-option project
				       :commit-privilege)
				     " before sending your message"
				   "\
, and type \\[patcher-mail-commit] to commit your project."))
				(otherwise
				 (patcher-error 'invalid-project-option
						:change-logs-appearance
						appearance)))))
	    ;; Manual mode. ChangeLogs are supposed to be written already so
	    ;; insertion can always be done right now.
	    (manual
	     ;; Without a specification, the diff is global so this is an
	     ;; opportunity to check for ChangeLogs consistency.
	     (unless (patcher-project-specification project)
	       (patcher-detect-inconsistent-change-logs project))
	     (case appearance
	       (verbatim
		(patcher-insert-diff project patcher-diff-marker)
		;; #### NOTE: when ChangeLog entries are part of the diff, we
		;; could try to convert the diff to a verbatim version instead
		;; of calling `patcher-insert-change-log-contents'.
		(patcher-insert-change-log-contents
		 project patcher-change-logs-marker))
	       (pack
		(patcher-detect-ephemeral-change-logs project)
		(let ((command (patcher-project-option project
				 :change-logs-diff-command)))
		  (cond (command
			 ;; The diff command is different. We need to rediff
			 ;; the ChangeLogs in all situations.
			 (patcher-insert-diff project patcher-diff-marker)
			 (patcher-diff-change-logs project)
			 (patcher-insert-diff project
			   patcher-change-logs-marker :change-logs))
			(t
			 ;; We use the same diff command:
			 (cond ((patcher-project-specification project)
				;; Some ChangeLogs may not be there. We must
				;; rediff them all.
				(patcher-insert-diff project
				  patcher-diff-marker)
				(patcher-diff-change-logs project)
				(patcher-insert-diff project
				  patcher-change-logs-marker :change-logs))
			       (t
				;; All ChangeLogs appear in the diff.
				(patcher-insert-diff project
				  patcher-diff-marker)
				(patcher-insert-diff project
				  patcher-change-logs-marker
				  :change-logs)))))))
	       (patch
		(patcher-detect-ephemeral-change-logs project)
		(cond ((patcher-project-specification project)
		       ;; Some ChangeLogs not be there. We must rediff the
		       ;; whole thing.
		       (patcher-diff-all project)
		       (patcher-insert-diff project patcher-diff-marker
					    :mixed))
		      (t
		       ;; All ChangeLogs appear in the diff.
		       (when (patcher-project-option project
			       :change-logs-diff-command)
			 (patcher-convert-change-logs-diff project))
		       (patcher-insert-diff project patcher-diff-marker
					    :mixed))))
	       ((nil)
		(patcher-insert-diff project patcher-diff-marker))
	       (otherwise
		(patcher-error 'invalid-project-option
			       :change-logs-appearance appearance)))
	     (patcher-message "\
To commit your changes, type \\[patcher-mail-commit]."))
	    (otherwise
	     (patcher-error 'invalid-project-option
			    :change-logs-updating updating)))))
    (diff
     (display-buffer (patcher-project-process-buffer patcher-project) t)
     (beep)
     (patcher-message "\
Error during diff. Type \\[patcher-mail-diff] to try again."))
    (change-logs-consistency
     (patcher-display-error-message
      (concat (patcher-inconsistent-change-logs-description
	       (nth 1 condition)
	       (nth 2 condition))
	      "\
\nPlease fix the problem and type \\[patcher-mail-diff] to try again.")))
    (sources-consistency
     (patcher-display-error-message
      (concat (patcher-inconsistent-sources-description
	       (nth 1 condition)
	       (nth 2 condition))
	      "\
\nPlease fix the problem and type \\[patcher-mail-diff] to try again.")))))



;; ===========================================================================
;; Patcher mail minor mode
;; ===========================================================================

(defun patcher-mail-change-subject ()
  "Read a new subject for the current project.

The new subject is propagated to all relevant buffers."
  (interactive)
  (patcher-change-subject patcher-project))

(defun patcher-mail-diff (&optional arg)
  "Regenerate the diff in the current Patcher mail buffer.

When called interactively, use a prefix to override the diff command
used for this project.

Note that this is *not* the way to specify files affected by this patch.
For that, either define a permanent subproject (see  `patcher-subprojects')
or call `patcher-mail' with a prefix argument."
  (interactive "P")
  (patcher-detect-undiffable-project patcher-project)
  (when (or (not (patcher-project-sources patcher-project))
	    (y-or-n-p "Really regenerate the diff ? "))
    (when arg
      (setf (patcher-project-diff-command patcher-project)
	    (read-shell-command "Diff command: "
				(patcher-project-diff-command
				 patcher-project))))
    (let ((change-log-buffers
	   (patcher-generated-change-logs patcher-project)))
      (when change-log-buffers
	(case (patcher-with-message
		  (format "\
Some ChangeLog skeletons for this patch have already been generated%s.
The relevant ChangeLog files are the following: %s.

Before regenerating the diff, please answer the question below to keep the
current skeletons (n), regenerate all of them (y) or choose interactively (i).

Beware that if you regenerate the skeletons, you will loose what you have
possibly already filled in."
		      (if (patcher-extent 'change-logs)
			  "\nand inserted into the current mail buffer"
			"")
		    (patcher-buffers-string change-log-buffers))
		(patcher-read-char "Regenerate ChangeLog skeletons? " "yni"))
	  (?y (patcher-ungenerate-change-logs patcher-project
		change-log-buffers))
	  (?i (patcher-ungenerate-change-logs patcher-project
		change-log-buffers
		:interactive t
		:prompt "Regenerate this skeleton? ")))))
    ;; #### NOTE: it is too complicated to decide whether or not to keep
    ;; inserted ChangeLog entries in the mail buffer here. For instance, the
    ;; user might have decided to keep the skeletons, but further modify them
    ;; after the diff. So let's just remove them.
    (patcher-delete-extent-and-region (patcher-extent 'change-logs))
    (patcher-delete-extent-and-region (patcher-extent 'diff))
    (patcher-diff-project patcher-project)))

(defun patcher-mail-insert-change-logs (&optional arg)
  "(Re)Insert ChangeLog entries in the current Patcher mail buffer.

When called interactively, use a prefix argument to temporarily override
the ChangeLogs appearance."
  (interactive "P")
  (let ((updating (patcher-project-option patcher-project
		    :change-logs-updating)))
    (patcher-condition-case condition
	(case updating
	  ((nil)
	   (patcher-error "This project does not use ChangeLogs"))
	  ((automatic manual)
	   (when (or (and (not (patcher-extent 'change-logs))
			  (not (patcher-change-logs)))
		     (y-or-n-p "\
ChangeLog entries already inserted.  Replace? "))
	     (let ((appearance (patcher-project-option patcher-project
				 :change-logs-appearance)))
	       (when (or (not appearance) arg)
		 (setq appearance
		       (let ((table '(("verbatim" . verbatim)
				      ("pack"     . pack)
				      ("patch"    . patch)
				      ("none"     . nil))))
			 (patcher-with-message (format "\
%sWhich type of appearance would you like to use?"
						   (if appearance
						       ""
						     "\
This project is set to not include ChangeLogs in mail buffers.\n"))
			   (cdr (assoc
				 (completing-read "\
Select a ChangeLog appearance (verbatim by default): "
				     table nil t nil nil "verbatim")
				 table))))))
	       (patcher-delete-extent-and-region (patcher-extent 'change-logs))
	       (patcher-within-extent (unused 'diff)
		 (dolist (extent (patcher-change-log-extents))
		   (patcher-delete-extent-and-region extent)))
	       (case appearance
		 ((nil))
		 (verbatim
		  (patcher-insert-change-log-contents
		   patcher-project patcher-change-logs-marker))
		 (pack
		  (patcher-detect-ephemeral-change-logs patcher-project)
		  (patcher-detect-undiffable-project patcher-project)
		  (patcher-diff-change-logs patcher-project)
		  (patcher-insert-diff patcher-project
		    patcher-change-logs-marker :change-logs))
		 (patch
		  (patcher-detect-ephemeral-change-logs patcher-project)
		  (patcher-detect-undiffable-project patcher-project)
		  (patcher-delete-extent-and-region (patcher-extent 'diff))
		  (patcher-diff-all patcher-project)
		  (patcher-insert-diff patcher-project patcher-diff-marker
				       :mixed))
		 (t
		  (patcher-error 'invalid-project-option
				 :change-logs-appearance appearance))))))
	  (t
	   (patcher-error 'invalid-project-option
			  :change-logs-updating updating)))
      (committed
       (patcher-display-error-message "\
This project has already been committed, so it is impossible to show the
ChangeLog entries as a diff because the diff would be empty."))
      (diff
       (display-buffer (patcher-project-process-buffer patcher-project) t)
       (beep)
       (patcher-message "\
Error during diff. Type \\[patcher-mail-insert-change-logs] to try again."))
      (change-logs-consistency
       (patcher-display-error-message
	(concat
	 (patcher-inconsistent-change-logs-description (nth 1 condition)
							(nth 2 condition))
	 "\
\nPlease fix the problem and type \\[patcher-mail-insert-change-logs] to try again.")))
      (sources-consistency
       (patcher-display-error-message
	(concat
	 (patcher-inconsistent-sources-description (nth 1 condition)
						    (nth 2 condition))
	 "\
\nPlease fix the problem and type \\[patcher-mail-insert-change-logs] to try again."))))))

(defun patcher-mail-first-change-log ()
  "Switch to first ChangeLog buffer."
  (interactive)
  (patcher-switch-to-first-change-log patcher-project))

(defun patcher-mail-last-change-log ()
  "Switch to the last ChangeLog buffer."
  (interactive)
  (patcher-switch-to-last-change-log patcher-project))

(defun patcher-mail-commit (&optional arg)
  "Prepare to, and possibly commit a change to a project's repository.
The change is the one that is announced in the mail buffer.

When called interactively, use a prefix (ARG) to override the commit
command to use.  Note that this is not meant to modify the source and
ChangeLog files affected by the commit: they are computed automatically."
  (interactive "P")
  (patcher-detect-committed-project patcher-project)
  (when arg
    (setf (patcher-project-commit-command patcher-project)
	  (read-shell-command "Commit command: "
			      (patcher-project-commit-command
			       patcher-project))))
  (setf (patcher-project-window-configuration patcher-project)
	(current-window-configuration))
  (let ((buffer (patcher-project-logmsg-buffer patcher-project)))
    (if buffer
	(erase-buffer buffer)
      (setq buffer
	    (setf (patcher-project-logmsg-buffer patcher-project)
		  (generate-new-buffer
		   (format "*%s Patcher Project Log Message*"
		       (patcher-project-name patcher-project)))))
      ;; Do it first! It kills local variables.
      (with-current-buffer buffer
	(patcher-logmsg-mode))
      (patcher-setup-auxiliary-buffer patcher-project buffer)))
  (if (not (patcher-project-option patcher-project :edit-log-message))
      (with-current-buffer (patcher-project-logmsg-buffer patcher-project)
	(patcher-logmsg-init)
	(patcher-condition-case nil
	    (patcher-logmsg-commit)
	  (commit
	   (display-buffer (patcher-project-process-buffer patcher-project) t)
	   (with-current-buffer (patcher-project-mail-buffer patcher-project)
	     (beep)
	     (patcher-message "\
Error during commit. Type \\[patcher-mail-commit] to try again.")))))
    (pop-to-buffer (patcher-project-logmsg-buffer patcher-project))
    (patcher-logmsg-init)))

(defun patcher-mail-kill ()
  "Kill the project related to the current mail buffer."
  (interactive)
  (patcher-kill-project patcher-project))

(defcustom patcher-mail-minor-mode-string " Patch"
  "*Patcher mail minor mode modeline string."
  :group 'patcher
  :type 'string)

(defcustom patcher-mail-minor-mode-hook nil
  "*Hooks to run after setting up Patcher mail minor mode."
  :group 'patcher
  :type 'hook)

(defvar patcher-mail-minor-mode-map
  (let ((map (make-sparse-keymap 'patcher-mail-minor-mode-map)))
    (define-key map [(control c) (control p) S] 'patcher-mail-change-subject)
    (define-key map [(control c) (control p) d] 'patcher-mail-diff)
    (define-key map [(control c) (control p) l]
      'patcher-mail-insert-change-logs)
    (define-key map [(control c) (control p) n] 'patcher-mail-first-change-log)
    (define-key map [(control c) (control p) p] 'patcher-mail-last-change-log)
    (define-key map [(control c) (control p) c] 'patcher-mail-commit)
    (define-key map [(control c) (control p) k] 'patcher-mail-kill)
    (define-key map [(control c) (control p) v] 'patcher-version)
    map)
  ;; Patcher minor mode keymap.
  )

(make-variable-buffer-local
 (defvar patcher-mail-minor-mode nil))

(defun patcher-insert-patcher-header ()
  ;; Insert a Patcher version header in the message.
  (save-excursion
    (goto-char (point-min))
    (unless (re-search-forward "^X-Generated-By: Patcher " nil t)
      ;; This search can fail in case of fake mail method.
      (when (re-search-forward
	     (concat "^" (regexp-quote mail-header-separator)) nil t)
	(goto-char (point-at-bol))
	(insert "X-Generated-By: " (patcher-version) "\n")))))

(defun patcher-mail-minor-mode (&optional arg)
  "Toggles Patcher mail minor mode.

Used for mails prepared with `patcher-mail'.  You're not supposed to use
this, unless you know what you're doing.

The Patcher Mail minor mode provides the following commands:
\\{patcher-mail-minor-mode-map}"
  (interactive "*P")
  (let ((was-off (not patcher-mail-minor-mode)))
    (setq patcher-mail-minor-mode
	  (if (null arg)
	      was-off
	    (> (prefix-numeric-value arg) 0)))
    (when (and patcher-mail-minor-mode was-off)
      (patcher-insert-patcher-header)
      (run-hooks 'patcher-mail-minor-mode-hook))))

(add-minor-mode 'patcher-mail-minor-mode
		patcher-mail-minor-mode-string
		patcher-mail-minor-mode-map)




;; ===========================================================================
;; Mail preparation routines
;; ===========================================================================

(defgroup patcher-mail nil
  "Patcher settings for mail buffers."
  :group 'patcher)

(defun patcher-before-send ()
  ;; Function hooked in the different mailing methods to perform some
  ;; checkings prior to sending the message.

  ;; #### NOTE: it is currently impossible (and probably not worth it) to
  ;; #### offer an automatic ChangeLog insertion or commit operation at that
  ;; #### point: we're already in an interactive call (the message sending
  ;; #### pocess) and a complex trickery would be necessary in case of
  ;; #### operation failure.  So it's simpler to just abort the sending, let
  ;; #### the user manually fix things, and re-send the message.

  ;; Check for a diff:
  (or (patcher-extent 'diff)
      (patcher-error "There's no diff in this message !"))
  ;; Check for ChangeLogs:
  (let ((check-insertion (patcher-project-option patcher-project
			   :check-change-logs-insertion)))
    (when (and check-insertion
	       (patcher-project-option patcher-project :change-logs-appearance)
	       (not (patcher-change-logs))
	       (or (eq check-insertion t)
		   ;; all other values are considered to be like 'ask
		   (not (y-or-n-p "\
You did not insert the ChangeLog entries.  Proceed with sending anyway ? "))))
      (patcher-error "\
Sending aborted.  Please insert the ChangeLogs first.")))
  ;; Check commit operation:
  (let ((check-commit (patcher-project-option patcher-project :check-commit)))
    (when (and check-commit
	       (patcher-project-option patcher-project :commit-privilege)
	       (not (patcher-project-committed-p patcher-project))
	       (or (eq check-commit t)
		   ;; all other values are considered to be like 'ask
		   (not (y-or-n-p "\
You did not commit your changes.  Proceed with sending anyway ? "))))
      (patcher-error "\
Sending aborted.  Please commit your changes first."))))

(defun patcher-after-send (&optional unused)
  ;; Function hooked in the different mailing methods to clean up the place
  ;; when a Patcher mail is sent.
  (patcher-delete-project patcher-project))

(defun patcher-install-send-hooks ()
  ;; Install before- and after-send hooks into the MUA.
  (cond ((eq major-mode 'mail-mode)
	 (add-local-hook 'mail-send-hook 'patcher-before-send)
	 (push '(patcher-after-send) mail-send-actions))
	((eq major-mode 'message-mode)
	 (add-local-hook 'message-send-hook 'patcher-before-send)
	 ;; `message-exit-actions' is probably more appropriate than
	 ;; `message-send-actions' to perform the cleanup.
	 (with-boundp 'message-exit-actions
	   (push '(patcher-after-send) message-exit-actions)))
	(t
	 (patcher-warning "\
Major mode: %s.
This mailing method is not fully supported by Patcher.
This is not critical though: Patcher won't be able to perform checks or
cleanups during mail sending.

Please report to <didier@xemacs.org>."
			  major-mode))))


;; Patcher FakeMail mode ====================================================

(defun patcher-fakemail-send ()
  "Pretend to send a fake Patcher mail.

Only perform the usual cleanup after real Patcher mails are sent."
  (interactive)
  (patcher-before-send)
  (patcher-after-send)
  (kill-buffer (current-buffer)))

(defvar patcher-fakemail-mode-map
  (let ((map (make-sparse-keymap 'patcher-fakemail-mode-map)))
    (define-key map [(control c) (control c)] 'patcher-fakemail-send)
    map))

(defun patcher-fakemail-mode ()
  "Sets up Patcher-FakeMail major mode.

Used for editing a fake Patcher mail.

The following command are available in a Fake Mail buffer:
\\{patcher-fakemail-mode-map}"
  (interactive)
  (kill-all-local-variables)
  (setq major-mode 'patcher-fakemail)
  (setq mode-name "Patcher-FakeMail")
  (use-local-map patcher-fakemail-mode-map)
  (run-hooks 'patcher-logmsg-mode-hook))


;; Interface to the different mailing methods ================================

(put 'patcher-with-mail-parameters 'lisp-indent-function 1)
(defmacro* patcher-with-mail-parameters (project &body body)
  ;; Wrap BODY in a let construct possibly defining user-full-name and
  ;; user-mail-address by Patcher options.
  ;; Return the value of BODY execution.
  ;; #### NOTE: why is it called like this ? Because I'm sure one day or
  ;; #### another, some sucker will ask for more parameters, like the mail
  ;; #### signature for instance ;-)
  `(let ((user-full-name (or (patcher-project-option ,project :user-name)
			     user-full-name))
	 (user-mail-address (or (patcher-project-option ,project :user-mail)
				user-mail-address)))
     ,@body))


(defun patcher-mail-compose-mail (project subject)
  "Prepare a patch-related mail with the `compose-mail' function.

This function uses the `:to-address' project option to determine the email
address for sending the message.  Otherwise, the address is prompted for.

See also the `mail-user-agent' variable."
  (patcher-with-mail-parameters project
    (compose-mail (or (patcher-project-option project :to-address)
		      (read-string "To address: "))
		  subject))
  (patcher-install-send-hooks))


(defun patcher-mail-sendmail (project subject)
  "Prepare a patch-related mail with the `mail' function.
This method requires the `sendmail' library.

This function uses the `:to-address' project option to determine the email
address for sending the message.  Otherwise, the address is prompted for."
  (require 'sendmail)
  (patcher-with-mail-parameters project
    (mail nil (or (patcher-project-option project :to-address)
		  (read-string "To address: "))
	  subject))
  (patcher-install-send-hooks))

(defun patcher-mail-message (project subject)
  "Prepare a patch-related mail with the `message-mail' function.
This method requires the `message' library.

This function uses the `:to-address' project option to determine the email
address for sending the message.  Otherwise, the address is prompted for."
  (require 'message)
  (patcher-with-mail-parameters project
    (message-mail (or (patcher-project-option project :to-address)
		      (read-string "To address: "))
		  subject))
  (patcher-install-send-hooks))

(defcustom patcher-mail-run-gnus 'prompt
  "*Whether Patcher should run Gnus.

The 'gnus mailing method of Patcher needs a running Gnus session.
If Gnus is not running at the time it is needed, Patcher can start
it (or not) depending on this variable:
- if nil, Patcher will abort execution,
- it 'prompt (the default), Patcher will ask you what to do,
- if t Patcher will unconditionally start Gnus.

See also the function `patcher-mail-gnus'."
  :group 'patcher-mail
  :type '(radio (const :tag "never" nil)
		(const :tag "ask user" prompt)
		(const :tag "as needed" t)))

(defcustom patcher-mail-run-gnus-other-frame t
  "*Whether Patcher should start Gnus in a new frame.

This is used in case Patcher has to start Gnus by itself \(see the
variable `patcher-mail-run-gnus').  Possible values are:
- nil:     start Gnus in the current frame,
- t:       start Gnus in a new frame,
- 'follow: start Gnus in a new frame, and also use this frame to prepare
	   the new Patcher message."
  :group 'patcher-mail
  :type '(radio (const :tag "Use current frame" nil)
		(const :tag "Create new frame" t)
		(const :tag "Create new frame, and use it for patcher"
		       follow)))

(defun patcher-mail-run-gnus ()
  ;; Start a gnus session.
  (require 'gnus)
  (save-excursion
    (cond ((eq patcher-mail-run-gnus-other-frame t)
	   (save-selected-frame (gnus-other-frame)))
	  ((eq patcher-mail-run-gnus-other-frame 'follow)
	   (gnus-other-frame))
	  ((not patcher-mail-run-gnus-other-frame)
	   (gnus))
	  (t
	   (patcher-error "\
Invalid value for `patcher-mail-run-gnus-other-frame': "
			  patcher-mail-run-gnus-other-frame)))))

(globally-declare-boundp 'gnus-article-copy)

(defun patcher-mail-gnus (project subject)
  "Prepare a patch-related mail with the `gnus-post-news' function.
Don't worry, this function can also send mails ;-).  This method
requires that you have Gnus *running* in your XEmacs session \(see
the variable `patcher-mail-run-gnus').

This function uses the `:gnus-group' project option to determine the Gnus
group to use \(as if you had typed `C-u a' on that group in the Group
buffer).  Otherwise, the group is prompted for."
  (require 'gnus-util)
  (with-fboundp '(gnus-alive-p gnus-post-news message-goto-body)
    (unless (gnus-alive-p)
      (cond ((not patcher-mail-run-gnus)
	     (patcher-error
	      "The 'gnus mailing method requires a running Gnus session"))
	    ((eq patcher-mail-run-gnus t)
	     (patcher-mail-run-gnus))
	    ((eq patcher-mail-run-gnus 'prompt)
	     (if (y-or-n-p "Gnus is not currently running.  Start it ? ")
		 (patcher-mail-run-gnus)
	       (patcher-error
		"The 'gnus mailing method requires a running Gnus session")))
	    (t
	     (patcher-error "Invalid value for `patcher-mail-run-gnus': "
			    patcher-mail-run-gnus))))
    (let ((gnus-newsgroup-name (or (patcher-project-option project
				     :gnus-group)
				   (read-string "Gnus group name: ")))
	  gnus-article-copy)
      (patcher-with-mail-parameters project
	(gnus-post-news 'post gnus-newsgroup-name)))
    (patcher-goto-subject)
    (insert subject)
    (message-goto-body)
    (patcher-install-send-hooks)))

(defun* patcher-mail-fake
    (project subject &aux (buffer (generate-new-buffer "*Patcher Fake Mail*")))
  "Prepare a patch-related fake mail.
Use this function if you want to do all that Patcher can do, apart from
sending a real mail.  This function generates a fake mail buffer which acts
as a standard Patcher mail buffer, apart from the fact that when you type
\\<patcher-fakemail-mode-map>\\[patcher-fakemail-send] in it, it doesn't
really send a mail, but just clean things up."
  (switch-to-buffer buffer)
  (insert "Subject: " subject "\n")
  (patcher-fakemail-mode))

(defun patcher-mail-setup (project)
  ;; Setup patcher-mail-minor-mode and initialize Patcher local variables in
  ;; mails (both generated or adapted).
  (setf (patcher-project-mail-buffer project) (current-buffer))
  (patcher-mail-minor-mode t)
  (cd (patcher-project-command-directory project))
  (setq patcher-project project))


;; Mail generation entry point ==============================================

;;;###autoload
(defun* patcher-mail (project)
  "Prepare a mail about a patch to apply on PROJECT.

When called interactively, prompt for a project name
\(see the variables `patcher-projects' and `patcher-subprojects') and
a subject for the mail.

With a additional prefix argument, create a temporary subproject by
prompting for an optional subdirectory and specific files as well. Files,
directories and even wildcards are acceptable in your specification.

With a prefix of 1, offer to relocate the project to another directory.
With a prefix of -1, do the same, but also create a temporary subproject of
the relocated project."
  (interactive (patcher-project-interactive current-prefix-arg))
  (funcall (intern (concat "patcher-mail-"
			   (symbol-name
			    (patcher-project-option project :mail-method t))))
	   project
	   (patcher-prefixed-subject project))
  (patcher-mail-setup project)
  (let ((mail-prologue (patcher-project-option project :mail-prologue)))
    (unless (zerop (length mail-prologue))
      (insert "\n" mail-prologue)))
  (save-excursion
    (insert "\n\n")
    (when (patcher-project-option project :change-logs-updating)
      (let ((appearance
	     (patcher-project-option project :change-logs-appearance)))
	(when (and appearance (not (eq appearance 'patch)))
	  (setq patcher-change-logs-marker (point-marker))
	  (insert "\n"))))
    (setq patcher-diff-marker (point-marker))
    (patcher-diff-project project)))


;; Mail adaptation entry point ==============================================

;;;###autoload
(defun patcher-mail-adapt (project)
  "Same as `patcher-mail', but for existing mails."
  (interactive (patcher-project-interactive current-prefix-arg))
  (patcher-goto-subject)
  (let ((beg (point))
	end
	old-subject)
    (while (progn (forward-line 1)
		  (looking-at "[ \t]")))
    (backward-char 1)
    (setq end (point))
    (skip-chars-backward " \t")
    (setq old-subject (buffer-substring beg (point)))
    (delete-region beg end)
    (while (string-match "\n[\t ]+" old-subject)
      (setq old-subject (replace-match " " t t old-subject)))
    (insert (patcher-prefixed-subject project old-subject)))
  (patcher-install-send-hooks)
  (patcher-mail-setup project)
  ;; #### NOTE: currently, I have simply discarded the mail-prologue
  ;; #### insertion for adapted mails. This is because mail adaptation is
  ;; #### mostly for replies in which you probably don't want the standard
  ;; #### prologue. However, this could be turned into a standard option.
  ;;  (let ((mail-prologue (patcher-project-option project :mail-prologue)))
  ;;    (unless (zerop (length mail-prologue))
  ;;      (insert "\n" mail-prologue)))
  (patcher-goto-signature)
  (when (patcher-project-option project :change-logs-updating)
    (let ((appearance
	   (patcher-project-option project :change-logs-appearance)))
      (when (and appearance (not (eq appearance 'patch)))
	(setq patcher-change-logs-marker (point-marker))
	(insert "\n"))))
  (setq patcher-diff-marker (point-marker))
  (patcher-diff-project project))




;; ==========================================================================
;; Gnus insinuation
;; ==========================================================================

;; Patcher Gnus Summary minor mode ==========================================

(defun patcher-gnus-summary-followup (&optional arg)
  "Prepare a Patcher followup from the Gnus Summary buffer.
See `patcher-mail' for more information."
  (interactive "P")
  (declare-fboundp (gnus-summary-followup nil))
  (call-interactively 'patcher-mail-adapt))

(defun patcher-gnus-summary-followup-with-original (&optional arg)
  "Prepare a Patcher followup from the Gnus Summary buffer.
The original message is yanked.
See `patcher-mail' for more information."
  (interactive "P")
  (declare-fboundp (gnus-summary-followup-with-original nil))
  (call-interactively 'patcher-mail-adapt))

(defun patcher-gnus-summary-reply (&optional arg)
  "Prepare a Patcher reply from the Gnus Summary buffer.
See `patcher-mail' for more information."
  (interactive "P")
  ;; #### NOTE: it is strange that this function's first argument is not
  ;; #### mandatory, as in the 3 other ones.
  (declare-fboundp (gnus-summary-reply))
  (call-interactively 'patcher-mail-adapt))

(defun patcher-gnus-summary-reply-with-original (&optional arg)
  "Prepare a Patcher reply from the Gnus Summary buffer.
The original message is yanked.
See `patcher-mail' for more information."
  (interactive "P")
  (declare-fboundp (gnus-summary-reply-with-original nil))
  (call-interactively 'patcher-mail-adapt))

(defcustom patcher-gnus-summary-minor-mode-string " Patch"
  "*Patcher Gnus Summary minor mode modeline string."
  :group 'patcher
  :type 'string)

(defcustom patcher-gnus-summary-minor-mode-hook nil
  "*Hooks to run after setting up Patcher Gnus Summary minor mode."
  :group 'patcher
  :type 'hook)

(defvar patcher-gnus-summary-minor-mode-map
  (let ((map (make-sparse-keymap 'patcher-gnus-summary-minor-mode-map)))
    (define-key map [(control c) (control p) f]
      'patcher-gnus-summary-followup)
    (define-key map [(control c) (control p) F]
      'patcher-gnus-summary-followup-with-original)
    (define-key map [(control c) (control p) r]
      'patcher-gnus-summary-reply)
    (define-key map [(control c) (control p) R]
      'patcher-gnus-summary-reply-with-original)
    map)
  ;; Patcher Gnus Summary minor mode keymap.
  )

(make-variable-buffer-local
 (defvar patcher-gnus-summary-minor-mode nil))

(defun patcher-gnus-summary-minor-mode (&optional arg)
  "Toggles Patcher Gnus Summary minor mode.

Used for Patcher messages composed as Gnus replies and followups.
You're not supposed to use this, unless you know what you're doing.

The Patcher Gnus Summary minor mode provides the following commands:
\\{patcher-gnus-summary-minor-mode-map}"
  (interactive "*P")
  (let ((was-off (not patcher-gnus-summary-minor-mode)))
    (setq patcher-gnus-summary-minor-mode
	  (if (null arg)
	      was-off
	    (> (prefix-numeric-value arg) 0)))
    (when (and patcher-gnus-summary-minor-mode was-off)
      (run-hooks 'patcher-gnus-summary-minor-mode-hook))))

(add-minor-mode
 'patcher-gnus-summary-minor-mode
 patcher-gnus-summary-minor-mode-string
 patcher-gnus-summary-minor-mode-map)


;; Patcher Gnus Article minor mode ==========================================

(defcustom patcher-gnus-article-minor-mode-string " Patch"
  "*Patcher Gnus Article minor mode modeline string."
  :group 'patcher
  :type 'string)

(defcustom patcher-gnus-article-minor-mode-hook nil
  "*Hooks to run after setting up Patcher Gnus Article minor mode."
  :group 'patcher
  :type 'hook)

(defvar patcher-gnus-article-minor-mode-map
  (let ((map (make-sparse-keymap 'patcher-gnus-article-minor-mode-map)))
    (define-key map [(control c) (control p) f]
      'patcher-gnus-summary-followup)
    (define-key map [(control c) (control p) F]
      'patcher-gnus-summary-followup-with-original)
    (define-key map [(control c) (control p) r]
      'patcher-gnus-summary-reply)
    (define-key map [(control c) (control p) R]
      'patcher-gnus-summary-reply-with-original)
    map)
  ;; Patcher Gnus Article minor mode keymap.
  )

(make-variable-buffer-local
 (defvar patcher-gnus-article-minor-mode nil))

(defun patcher-gnus-article-minor-mode (&optional arg)
  "Toggles Patcher Gnus Article minor mode.

Used for Patcher messages composed as Gnus replies and followups.
You're not supposed to use this, unless you know what you're doing.

The Patcher Gnus Article minor mode provides the following commands:
\\{patcher-gnus-article-minor-mode-map}"
  (interactive "*P")
  (let ((was-off (not patcher-gnus-article-minor-mode)))
    (setq patcher-gnus-article-minor-mode
	  (if (null arg)
	      was-off
	    (> (prefix-numeric-value arg) 0)))
    (when (and patcher-gnus-article-minor-mode was-off)
      (run-hooks 'patcher-gnus-article-minor-mode-hook))))

(add-minor-mode
 'patcher-gnus-article-minor-mode
 patcher-gnus-article-minor-mode-string
 patcher-gnus-article-minor-mode-map)


;; Insinuation ==============================================================

;;;###autoload
(defun patcher-insinuate-gnus ()
  "Hook Patcher functionality into Gnus.

This function should be called from your gnusrc file."
  (add-hook 'gnus-summary-mode-hook
	    (lambda () (patcher-gnus-summary-minor-mode t)))
  (add-hook 'gnus-article-mode-hook
	    (lambda () (patcher-gnus-article-minor-mode t))))


(provide 'patcher-mail)

;;; patcher-mail.el ends here
